package sos.util;

import java.io.File;
import java.io.RandomAccessFile;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import java.lang.reflect.Method;


/**
 * This class contains static methods for several file operations
 * 
 * @author Florian Schreiber <fs@sos-berlin.com>
 * @since 2006-10-10
 * @version 2007-06-25
 */

public class SOSFileOperations {
	
	/**
	 * copyFile and renameFile create the target directory if it does not yet exist.
	 */
	public static final int CREATE_DIR = 0x01;
	
	/**
	 * Disables error messages caused by specified (source and/or target) but not existing files.
	 */
	public static final int GRACIOUS = 0x02;
	
	/**
	 * Denies overwriting of already existing files.<br/>
	 * By default files are overwritten.
	 */
	public static final int NOT_OVERWRITE = 0x04;
	
	/**
	 * Enables recursive processing of all subdirectories.<br/>
	 * It only Takes effect if the method's specified source is a directory.
	 */
	public static final int RECURSIVE = 0x08;
	
	/**
	 * Forces removeFile to include directories for removal.
	 */
	public static final int REMOVE_DIR = 0x10;
	
	/**
	 * Completely overwrites data when deleting files.
	 */
	public static final int WIPE = 0x20;
	
	
	// bildet von diversen bool-aeqivalenten Strings auf "true" bzw. "false" ab
	private static Hashtable BOOL = new Hashtable();
	
	static {
		
		SOSFileOperations.BOOL.put("true", "true");
		SOSFileOperations.BOOL.put("false", "false");
		SOSFileOperations.BOOL.put("j", "true");
		SOSFileOperations.BOOL.put("ja", "true");
		SOSFileOperations.BOOL.put("y", "true");
		SOSFileOperations.BOOL.put("yes", "true");
		SOSFileOperations.BOOL.put("n", "false");
		SOSFileOperations.BOOL.put("nein", "false");
		SOSFileOperations.BOOL.put("no", "false");
		SOSFileOperations.BOOL.put("on", "true");
		SOSFileOperations.BOOL.put("off", "false");
		SOSFileOperations.BOOL.put("1", "true");
		SOSFileOperations.BOOL.put("0", "false");
		SOSFileOperations.BOOL.put("all", "true");
		SOSFileOperations.BOOL.put("none", "false");
	}
	
	/**
	 * Maps the strings "true","j","ja","y","yes","on","1" and "all" to true.<br/>
	 * Maps the strings "false","n","nein","no","off","0" and "none" to false.<br/>
	 * The mapping is case insensitive, e.g. the string <em>"nEIN"</em> is mapped to <em>false</em>.<br/>
	 * If the value cannot be evaluated to a boolean then an exception is thrown.
	 * 
	 * @param value
	 * @return true oder false
	 */
	public static boolean toBoolean(String value) throws Exception {
		
		try {

			if (value == null)
				throw new Exception("null");

			String v = value.toLowerCase();
			String bool = (String) SOSFileOperations.BOOL.get(v);

			if (bool == null)
				throw new Exception("\"" + value + "\"");

			return bool.equals("true");

		} catch (Exception e) {
			throw new Exception("cannot evaluate to boolean: " + e.getMessage());
		}
	}
	
	
	/**
	 * Checks for existence of file or directory.
	 * 
	 * @param file
	 * 					file or directory
	 * @param logger
	 * 					SOSLogger
	 * @return true if file exists or false if not
	 * @throws Exception
	 */
	public static boolean existsFile(String file, SOSLogger logger) throws Exception {
		
		File filename = new File(file);
		return existsFile(filename, null, 0, logger);
	}
	
	/**
	 * Checks for file existence.
	 * if file is a directory and <em>fileSpec</em> is not NULL
	 * <em>fileSpec</em> is applied for matching files in the directory.
	 * In this case true will only be returned if at least one file was matched.
	 * 
	 * @param file
	 * 					file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param logger
	 * 					SOSLogger
	 * @return true if file exists or false if not
	 * @throws Exception
	 */
	public static boolean existsFile(String file, String fileSpec, SOSLogger logger) throws Exception {
		
		File filename = new File(file);
		return existsFile(filename, fileSpec, Pattern.CASE_INSENSITIVE, logger);
	}
	
	/**
	 * Checks for file existence.
	 * if file is a directory and <em>fileSpec</em> is not NULL
	 * <em>fileSpec</em> is applied for matching files in the directory.
	 * In this case true will only be returned if at least one file was matched.
	 * 
	 * @param file
	 * 					file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param fileSpecFlags
	 *          Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see
	 * 				<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param logger
	 * 					SOSLogger
	 * @return true if file exists or false if not
	 * @throws Exception
	 */
	public static boolean existsFile(String file, String fileSpec, int fileSpecFlags, SOSLogger logger) throws Exception {
		
		File filename = new File(file);
		return existsFile(filename, fileSpec, fileSpecFlags, logger);
	}
	
	/**
	 * Checks for existence of file or directory.
	 * 
	 * @param file
	 * 					file or directory
	 * @param logger
	 * 					SOSLogger
	 * @return true if file exists or false if not
	 * @throws Exception
	 */
	public static boolean existsFile(File file, SOSLogger logger) throws Exception {
		
		return existsFile(file, null, 0, logger);
	}
	
	/**
	 * Checks for file existence.
	 * if file is a directory and <em>fileSpec</em> is not NULL
	 * <em>fileSpec</em> is applied for matching files in the directory.
	 * In this case true will only be returned if at least one file was matched.
	 * 
	 * @param file
	 * 					file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param logger
	 * 					SOSLogger
	 * @return true if file exists or false if not
	 * @throws Exception
	 */
	public static boolean existsFile(File file, String fileSpec, SOSLogger logger) throws Exception {
		
		return existsFile(file, fileSpec, Pattern.CASE_INSENSITIVE, logger);
	}
	
	/**
	 * Checks for file existence.
	 * if file is a directory and <em>fileSpec</em> is not NULL
	 * <em>fileSpec</em> is applied for matching files in the directory.
	 * In this case true will only be returned if at least one file was matched.
	 * 
	 * @param file
	 * 					file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param fileSpecFlags
	 *          Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see
	 * 				<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param logger
	 * 					SOSLogger
	 * @return true if file exists or false if not
	 * @throws Exception
	 */
	public static boolean existsFile(File file, String fileSpec, int fileSpecFlags, SOSLogger logger) throws Exception {
		
		try {
			
			log_debug1("arguments for existsFile:", logger);
			log_debug1("argument file=" + file.toString(), logger);
			log_debug1("argument fileSpec=" + fileSpec, logger);
			String msg = "";
			if (has(fileSpecFlags, Pattern.CANON_EQ)) msg += "CANON_EQ";
			if (has(fileSpecFlags, Pattern.CASE_INSENSITIVE)) msg += "CASE_INSENSITIVE";
			if (has(fileSpecFlags, Pattern.COMMENTS)) msg += "COMMENTS";
			if (has(fileSpecFlags, Pattern.DOTALL)) msg += "DOTALL";
			if (has(fileSpecFlags, Pattern.MULTILINE)) msg += "MULTILINE";
			if (has(fileSpecFlags, Pattern.UNICODE_CASE)) msg += "UNICODE_CASE";
			if (has(fileSpecFlags, Pattern.UNIX_LINES)) msg += "UNIX_LINES";
			log_debug1("argument fileSpecFlags=" + msg, logger);
			
			
			String filename = file.getPath();
			
			filename = substituteAllDate(filename);
			
			// should any opening and closing brackets be found in the file name, then this is an error
			Matcher m = Pattern.compile("\\[[^]]*\\]").matcher(filename);
			if ( m.find() ) throw new Exception("unsupported file mask found: "+ m.group());
			
			file = new File(filename);
			
			if ( !file.exists() ) {
				log("checking file "+ file.getAbsolutePath() +": no such file or directory", logger);
				return false;
				
			} else {
				
				// Es ist eine Datei und sie existiert
				if ( !file.isDirectory() ) {
					
					log("checking file "+ file.getCanonicalPath() +": file exists", logger);
					return true;
					
				// Es ist ein Verzeichnis und es existiert
				} else {
					
					// wenn kein fileSpec angegeben wurde, reicht die simple Existenz aus
					if ( fileSpec==null || fileSpec.length()==0 ) {
						log("checking file "+ file.getCanonicalPath() +": directory exists", logger);
						return true;
					}
					
					// Ein Dateifilter wurde angegeben
					Vector filelist = SOSFile.getFilelist(file.getPath(), fileSpec, fileSpecFlags);
					
					if (filelist.isEmpty()) {
						log("checking file "+ file.getCanonicalPath() +": directory contains no files matching "+ fileSpec, logger);
						return false;
						
					} else {
						log("checking file "+ file.getCanonicalPath() +": directory contains "+ filelist.size() +" file(s) matching "+ fileSpec, logger);
						for (int i = 0; i < filelist.size(); i++)
							log("found "+ ((File) filelist.get(i)).getCanonicalPath(), logger);
						return true;
					}
				}
			}
			
		} catch (Exception e) {
			throw new Exception("error checking file: "+ file.toString() +": "+ e.getMessage());
		}
	}
	
	
	/**
	 * Removes one file or all files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory.
	 * If <em>file</em> is a directory all contained files are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(File file, SOSLogger logger) throws Exception {

		return removeFile(file, ".*", 0, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or all files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory.
	 * If <em>file</em> is a directory all contained files are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #GRACIOUS}, {@link #RECURSIVE}, {@link #REMOVE_DIR} and {@link #WIPE} 
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(File file, int flags, SOSLogger logger) throws Exception {

		return removeFile(file, ".*", flags, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or several files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory (<em>fileSpec</em> is ignored).
	 * If <em>file</em> is a directory all contained files matching the expression <em>fileSpec</em> are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(File file, String fileSpec, SOSLogger logger) throws Exception {

		return removeFile(file, fileSpec, 0, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or several files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory (<em>fileSpec</em> is ignored).
	 * If <em>file</em> is a directory all contained files matching the expression <em>fileSpec</em> are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #GRACIOUS}, {@link #RECURSIVE}, {@link #REMOVE_DIR} and {@link #WIPE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(File file, String fileSpec, int flags, SOSLogger logger) throws Exception {

		return removeFile(file, fileSpec, flags, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or several files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory (<em>fileSpec</em> is ignored).
	 * If <em>file</em> is a directory all contained files matching the expression <em>fileSpec</em> are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #GRACIOUS}, {@link #RECURSIVE}, {@link #REMOVE_DIR} and {@link #WIPE}
	 * @param fileSpecFlags
	 *          Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see
	 * 				<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(File file, String fileSpec, int flags, int fileSpecFlags, SOSLogger logger) throws Exception {
		
		boolean gracious;
		boolean recursive;
		boolean remove_dir;
		boolean wipe;
		
		int nrOfRemovedFiles = 0;
		int nrOfRemovedDirectories = 0;
		
		try {
			
			recursive = has(flags, SOSFileOperations.RECURSIVE);
			gracious = has(flags, SOSFileOperations.GRACIOUS);
			wipe = has(flags, SOSFileOperations.WIPE);
			remove_dir = has(flags, SOSFileOperations.REMOVE_DIR);
			
			// Argumente ausgeben
			log_debug1("arguments for removeFile:", logger);
			log_debug1("argument file=" + file, logger);
			log_debug1("argument fileSpec=" + fileSpec, logger);
			
			String msg = "";
			if (has(flags, SOSFileOperations.GRACIOUS))
				msg += "GRACIOUS ";
			if (has(flags, SOSFileOperations.RECURSIVE))
				msg += "RECURSIVE ";
			if (has(flags, SOSFileOperations.REMOVE_DIR))
				msg += "REMOVE_DIR ";
			if (has(flags, SOSFileOperations.WIPE))
				msg += "WIPE ";
			log_debug1("argument flags=" + msg, logger);

			msg = "";
			if (has(fileSpecFlags, Pattern.CANON_EQ))
				msg += "CANON_EQ ";
			if (has(fileSpecFlags, Pattern.CASE_INSENSITIVE))
				msg += "CASE_INSENSITIVE ";
			if (has(fileSpecFlags, Pattern.COMMENTS))
				msg += "COMMENTS ";
			if (has(fileSpecFlags, Pattern.DOTALL))
				msg += "DOTALL ";
			if (has(fileSpecFlags, Pattern.MULTILINE))
				msg += "MULTILINE ";
			if (has(fileSpecFlags, Pattern.UNICODE_CASE))
				msg += "UNICODE_CASE ";
			if (has(fileSpecFlags, Pattern.UNIX_LINES))
				msg += "UNIX_LINES ";
			log_debug1("argument fileSpecFlags=" + msg, logger);
			
			if ( !file.exists() ) {
				if ( gracious ) {
					log("cannot remove file: file does not exist: "+ file.getCanonicalPath(), logger);
					return false;
				} else
					throw new Exception("file does not exist: " + file.getCanonicalPath());
			}
			
			Vector fileList;
			
			// Verzeichnis
			if ( file.isDirectory() ) {
				if ( !file.canRead() )
					throw new Exception("directory is not readable: "+ file.getCanonicalPath());
				
				log("remove ["+ fileSpec +"] from "+ file.getCanonicalPath() + (recursive ? " (recursive)" : ""), logger);
				fileList = SOSFile.getFilelist(file.getPath(), fileSpec, fileSpecFlags, recursive);
			}
			// einzelne Datei
			else {
				fileList = new Vector();
				fileList.add(file);
			}
			
			File currentFile;
			for (int i = 0; i < fileList.size(); i++) {
				
				currentFile = (File) fileList.get(i);
				log("remove file " + currentFile.getCanonicalPath(), logger);
				
				if (wipe) {
					if ( !wipe(currentFile, logger) )
						throw new Exception("cannot remove file: "+ currentFile.getCanonicalPath());
				}
				else {
					if ( !currentFile.delete() )
						throw new Exception("cannot remove file: "+ currentFile.getCanonicalPath());
				}
				
				nrOfRemovedFiles++;
			}
			
			// sollen auch Verzeichnisse entfernt werden ?
			if ( remove_dir ) {
				
				int firstSize = SOSFile.getFolderlist(file.getPath(), ".*", 0, recursive).size();
				
				if ( recursive ) {
					recDeleteEmptyDir(file, fileSpec, fileSpecFlags, logger);
				}
				else {
					Vector list = SOSFile.getFolderlist(file.getPath(), fileSpec, fileSpecFlags);
					File f;
					
					for (int i=0; i<list.size(); i++) {
						f = (File)list.get(i);
						if ( f.isDirectory() ) {
							if ( !f.canRead() ) throw new Exception("directory is not readable: "+ f.getCanonicalPath());
							
							if ( f.list().length==0 ) {
								if ( !f.delete() ) throw new Exception("cannot remove dirctory: "+f.getCanonicalPath());
								log("remove directory "+f.getPath(), logger);
							}
						}
					}
				}
				
				nrOfRemovedDirectories = firstSize - SOSFile.getFolderlist(file.getPath(), ".*", 0, recursive).size();
			}
			
			msg = "";
			if ( remove_dir ) {
				if ( nrOfRemovedDirectories==1 )
					msg = " + 1 directory removed";
				else
					msg = " + "+ nrOfRemovedDirectories +" directories removed";
			}
			
			log(nrOfRemovedFiles + " file(s) removed"+msg, logger);
			
			return (nrOfRemovedFiles > 0);
			
		} catch (Exception e) {
			throw new Exception("error occurred removing file(s): " + e.getMessage());
		}
	}
	
	private static boolean recDeleteEmptyDir(File dir, String fileSpec, int fileSpecFlags, SOSLogger logger) throws Exception {
		
		if ( dir.isDirectory() ) {
			if ( !dir.canRead() )
				throw new Exception("directory is not readable: "+ dir.getCanonicalPath());
		}
		else {
			// eine Datei gleich ausschlieen
			return false;
		}
		
		// Verzeichnisinhalt inklusive Verzeichnisse einlesen - nicht rekursiv!
		File[] list = dir.listFiles();
		if ( list.length==0 ) return true;
		
		Pattern p = Pattern.compile(fileSpec, fileSpecFlags);
		
		File f;
		for ( int i=0; i<list.length; i++ ) {
			f = list[i];
			
			if ( recDeleteEmptyDir(f, fileSpec, fileSpecFlags, logger) ) {
				if ( p.matcher(f.getName()).matches() ) {
					if ( !f.delete() ) throw new Exception("cannot remove dirctory: "+f.getCanonicalPath());
					log("remove directory "+f.getPath(), logger);
				}
			}
		}
		
		// ist das Verzeichnnis jetzt leer?
		return ( dir.list().length==0 );
	}
	
	
	/**
	 * Removes one file or all files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory.
	 * If <em>file</em> is a directory all contained files are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(String file, SOSLogger logger) throws Exception {

		return removeFile(new File(file), ".*", 0, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or all files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory.
	 * If <em>file</em> is a directory all contained files are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #GRACIOUS}, {@link #RECURSIVE}, {@link #REMOVE_DIR} and {@link #WIPE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(String file, int flags, SOSLogger logger) throws Exception {

		return removeFile(new File(file), ".*", flags, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or several files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory (<em>fileSpec</em> is ignored).
	 * If <em>file</em> is a directory all contained files matching the expression <em>fileSpec</em> are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(String file, String fileSpec, SOSLogger logger) throws Exception {

		return removeFile(new File(file), fileSpec, 0, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or several files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory (<em>fileSpec</em> is ignored).
	 * If <em>file</em> is a directory all contained files matching the expression <em>fileSpec</em> are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #GRACIOUS}, {@link #RECURSIVE}, {@link #REMOVE_DIR} and {@link #WIPE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(String file, String fileSpec, int flags, SOSLogger logger) throws Exception {

		return removeFile(new File(file), fileSpec, flags, Pattern.CASE_INSENSITIVE, logger);
	}

	/**
	 * Removes one file or several files of a directory from the file system.
	 * Removes one file if <em>file</em> is a file but no directory (<em>fileSpec</em> is ignored).
	 * If <em>file</em> is a directory all contained files matching the expression <em>fileSpec</em> are removed.
	 * 
	 * @param file
	 *          file or directory
	 * @param fileSpec
	 * 					Regular expression for file filtering if file is a directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #GRACIOUS}, {@link #RECURSIVE}, {@link #REMOVE_DIR} and {@link #WIPE}
	 * @param fileSpecFlags
	 *          Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see
	 * 				<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean removeFile(String file, String fileSpec, int flags, int fileSpecFlags, SOSLogger logger) throws Exception {

		return removeFile(new File(file), fileSpec, flags, fileSpecFlags, logger);
	}
	
	
	/**
	 * Copies one file or all files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(File source, File target, SOSLogger logger) throws Exception {

		return copyFile(source, target, ".*", 0, Pattern.CASE_INSENSITIVE, null, null,
				logger);
	}
	
	/**
	 * Copies one file or all files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(File source, File target, int flags, SOSLogger logger) throws Exception {

		return copyFile(source, target, ".*", flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(File source, File target, String fileSpec, SOSLogger logger) throws Exception {

		return copyFile(source, target, fileSpec, 0, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(File source, File target, String fileSpec,
			int flags, SOSLogger logger) throws Exception {

		return copyFile(source, target, fileSpec, flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory, this argument is also used as target
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(File source, String fileSpec, int flags,
			int fileSpecFlags, String replacing, String replacement, SOSLogger logger)
			throws Exception {

		String mode = "copy";
		return transferFile(source, null, fileSpec, flags, fileSpecFlags, replacing, replacement, mode, logger);
	}
	
	/**
	 * Copies one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(File source, File target, String fileSpec,
			int flags, int fileSpecFlags, String replacing, String replacement,
			SOSLogger logger) throws Exception {

		String mode = "copy";
		return transferFile(source, target, fileSpec, flags, fileSpecFlags, replacing, replacement, mode, logger);
	}

	/**
	 * Copies one file or all files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(String source, String target, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return copyFile(sourceFile, targetFile, ".*", 0, Pattern.CASE_INSENSITIVE, null,
				null, logger);
	}

	/**
	 * Copies one file or all files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(String source, String target, int flags, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return copyFile(sourceFile, targetFile, ".*", flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(String source, String target, String fileSpec, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return copyFile(sourceFile, targetFile, fileSpec, 0, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(String source, String target, String fileSpec,
			int flags, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return copyFile(sourceFile, targetFile, fileSpec, flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(String source, String target, String fileSpec,
			int flags, int fileSpecFlags, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return copyFile(sourceFile, targetFile, fileSpec, flags, fileSpecFlags, null, null, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory, this argument is also used as target
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(String source, String fileSpec, int flags,
			int fileSpecFlags, String replacing, String replacement, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		return copyFile(sourceFile, null, fileSpec, flags, fileSpecFlags, replacing, replacement, logger);
	}

	/**
	 * Copies one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean copyFile(String source, String target, String fileSpec,
			int flags, int fileSpecFlags, String replacing, String replacement, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return copyFile(sourceFile, targetFile, fileSpec, flags, fileSpecFlags, replacing, replacement, logger);
	}
	
	
	/**
	 * Renames/moves one file or all files of a directory.
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, SOSLogger logger)
			throws Exception {

		return renameFile(source, target, ".*", 0, Pattern.CASE_INSENSITIVE, null, null, logger);
	}
	
	/**
	 * Renames/moves one file or all files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, int flags,
			SOSLogger logger) throws Exception {

		return renameFile(source, target, ".*", flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}
	
	/**
	 * Renames/moves one file or several files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, String fileSpec,
			SOSLogger logger) throws Exception {

		return renameFile(source, target, fileSpec, 0, Pattern.CASE_INSENSITIVE, null, null, logger);
	}
	
	/**
	 * Renames/moves one file or several files of a directory.
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, String fileSpec,
			int flags, SOSLogger logger) throws Exception {

		return renameFile(source, target, fileSpec, flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, String fileSpec,
			int flags, int fileSpecFlags, SOSLogger logger) throws Exception {

		return renameFile(source, target, fileSpec, flags, fileSpecFlags, null, null, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, String fileSpec,
			String replacing, String replacement, SOSLogger logger) throws Exception {

		return renameFile(source, target, fileSpec, 0, Pattern.CASE_INSENSITIVE, replacing, replacement, logger);
	}
	
	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, String fileSpec,
			int flags, String replacing, String replacement, SOSLogger logger)
			throws Exception {

		return renameFile(source, target, fileSpec, flags, Pattern.CASE_INSENSITIVE, replacing, replacement, logger);
	}
	
	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory, this argument is also used as target
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, String fileSpec, int flags,
			int fileSpecFlags, String replacing, String replacement, SOSLogger logger)
			throws Exception {
		return renameFile(source, null, fileSpec, flags, fileSpecFlags, replacing, replacement, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(File source, File target, String fileSpec,
			int flags, int fileSpecFlags, String replacing, String replacement,
			SOSLogger logger) throws Exception {

		String mode = "rename";
		return transferFile(source, target, fileSpec, flags, fileSpecFlags, replacing, replacement, mode, logger);
	}

	/**
	 * Renames/moves one file or all files of a directory. 
	 * Already existing files are overwritten
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, SOSLogger logger)
			throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, ".*", 0, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Renames/moves one file or all files of a directory. 
	 * By default already existing files are overwritten
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, int flags,
			SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, ".*", flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, String fileSpec,
			SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, fileSpec, 0, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, String fileSpec,
			int flags, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, fileSpec, flags, Pattern.CASE_INSENSITIVE, null, null, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, String fileSpec,
			int flags, int fileSpecFlags, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, fileSpec, flags, fileSpecFlags, null, null, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * Already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, String fileSpec,
			String replacing, String replacement, SOSLogger logger) throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, fileSpec, 0, Pattern.CASE_INSENSITIVE, replacing, replacement, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, String fileSpec,
			int flags, String replacing, String replacement, SOSLogger logger)
			throws Exception {

		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, fileSpec, flags, Pattern.CASE_INSENSITIVE, replacing, replacement, logger);
	}

	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory, this argument is also used as target
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String fileSpec, int flags,
			int fileSpecFlags, String replacing, String replacement, SOSLogger logger)
			throws Exception {

		File sourceFile = new File(source);
		return renameFile(sourceFile, null, fileSpec, flags, fileSpecFlags, replacing, replacement, logger);
	}
	
	/**
	 * Renames/moves one file or several files of a directory. 
	 * By default already existing files are overwritten.
	 * 
	 * @param source
	 *          Source file or source dirctory
	 * @param target
	 *          Target file or target directory
	 * @param fileSpec
	 *          Regular expression - file filter, used if source is a directory 
	 * @param flags
	 * 					Bitmask for following flags: {@link #CREATE_DIR}, {@link #GRACIOUS}, {@link #NOT_OVERWRITE} and {@link #RECURSIVE}
	 * @param fileSpecFlags
	 * 					Pattern bitmask providing the regular expression <em>fileSpec</em>
	 * @see 
	 * 					<a href="http://java.sun.com/j2se/1.4.2/docs/api/java/util/regex/Pattern.html">java.util.regex.Pattern</a>
	 * @param replacing
	 * 					Regular expression for file name replacement with <em>replacement</em>.
	 * 					<em>replacing</em> is ignored if it is NULL.
	 * 					Requires the parameter <em>replacement</em> being not NULL. 
	 * @param replacement
	 * 					String for replacement of the expression replacing.
	 * 					Replacements for multiple capturing groups are separated by a semicolon ";"					
	 * 					<em>replacement</em> is ignored if it is NULL.
	 *					Requires the parameter <em>replacing</em> being not NULL.
	 * @param logger
	 * 					SOSLogger
	 * @throws Exception
	 */
	public static boolean renameFile(String source, String target, String fileSpec,
			int flags, int fileSpecFlags, String replacing, String replacement,
			SOSLogger logger) throws Exception {
		
		File sourceFile = new File(source);
		File targetFile = (target == null) ? null : new File(target);
		return renameFile(sourceFile, targetFile, fileSpec, flags, fileSpecFlags, replacing, replacement, logger);
	}

	
	/**
	 * Used for copyFile and renameFile
	 */
	private static boolean transferFile(File source, File target, String fileSpec, int flags, int fileSpecFlags,
			String replacing, String replacement, String mode, SOSLogger logger) throws Exception {
		
		int nrOfTransferedFiles = 0;
		boolean create_dir;
		boolean gracious;
		boolean overwrite;
		boolean replace = false;
		
		// modi
		boolean copying = false;
		boolean renaming = false;
		
		String targetFilename;
		
		try {
			
			if (mode.equals("copy"))
				copying = true;
			else if (mode.equals("rename"))
				renaming = true;
			else
				throw new Exception("unsupported mode: " + mode);
			
			// welche Flags sind gesetzt?
			create_dir = has(flags, SOSFileOperations.CREATE_DIR);
			gracious = has(flags, SOSFileOperations.GRACIOUS);
			overwrite = !has(flags, SOSFileOperations.NOT_OVERWRITE);
			
			// Argumente ausgeben
			if ( copying )
				log_debug1("arguments for copyFile:", logger);
			else if ( renaming )
				log_debug1("arguments for renameFile:", logger);
			
			log_debug1("argument source=" + source.toString(), logger);
			if ( target!=null ) log_debug1("argument target=" + target.toString(), logger);
			log_debug1("argument fileSpec=" + fileSpec, logger);
			
			String msg = "";
			if (has(flags, SOSFileOperations.CREATE_DIR)) msg += "CREATE_DIR ";
			if (has(flags, SOSFileOperations.GRACIOUS)) msg += "GRACIOUS ";
			if (has(flags, SOSFileOperations.NOT_OVERWRITE)) msg += "NOT_OVERWRITE ";
			if (has(flags, SOSFileOperations.RECURSIVE)) msg += "RECURSIVE ";
			log_debug1("argument flags=" + msg, logger);
			
			msg = "";
			if (has(fileSpecFlags, Pattern.CANON_EQ)) msg += "CANON_EQ ";
			if (has(fileSpecFlags, Pattern.CASE_INSENSITIVE)) msg += "CASE_INSENSITIVE ";
			if (has(fileSpecFlags, Pattern.COMMENTS)) msg += "COMMENTS ";
			if (has(fileSpecFlags, Pattern.DOTALL)) msg += "DOTALL ";
			if (has(fileSpecFlags, Pattern.MULTILINE)) msg += "MULTILINE ";
			if (has(fileSpecFlags, Pattern.UNICODE_CASE)) msg += "UNICODE_CASE ";
			if (has(fileSpecFlags, Pattern.UNIX_LINES)) msg += "UNIX_LINES ";
			log_debug1("argument fileSpecFlags=" + msg, logger);
			
			log_debug1("argument replacing=" + replacing, logger);
			log_debug1("argument replacement=" + replacement, logger);

			// Wenn ersetzt werden soll, ist es gltig?
			if (replacing != null || replacement != null) {

				if (replacing == null)
					throw new Exception("replacing cannot be null if replacement is set");
				if (replacement == null)
					throw new Exception("replacement cannot be null if replacing is set");

				if (!replacing.equals("")) {
					try {
						Pattern.compile(replacing); // Pattern validieren
					} catch (PatternSyntaxException pse) {
						throw new Exception("invalid pattern '" + replacing + "'");
					}
					replace = true;
				}
			}
			
			// Existiert die Quelle?
			if ( !source.exists() ) {
				if (gracious) {
					log(nrOfTransferedFiles + " file(s) renamed", logger);
					return (nrOfTransferedFiles > 0);
				} else
					throw new Exception("file or directory does not exist: "+ source.getCanonicalPath());
			}
			
			if ( !source.canRead() )
				throw new Exception("file or directory is not readable: "+ source.getCanonicalPath());
			
			
			// Target
			// Substitution im Zielverzeichnisnamen
			if ( target!=null ) {
				targetFilename = substituteAllDate( target.getPath() );
				targetFilename = substituteAllDirectory(targetFilename, source.getPath());
				
				// should any opening and closing brackets be found in the file name, then this is an error
				Matcher m = Pattern.compile("\\[[^]]*\\]").matcher(targetFilename);
				if ( m.find() ) throw new Exception("unsupported file mask found: "+ m.group());
				
				// zum Schluss nderungen zuweisen
				target = new File(targetFilename);
			}
			
			
			// optional Zielverzeichnis anlegen
			if ( create_dir ) {
				if ( target!=null && !target.exists() ) {
					if ( target.mkdirs() )
						log("create target directory " + target.getCanonicalPath(), logger);
					else
						throw new Exception("cannot create directory "+target.getCanonicalPath());
				}
			}
			
			
			// Liste der Dateien
			Vector list = null;
			
			// 1. Quelle ist ein Verzeichnis
			if (source.isDirectory()) {
				if (target != null) {
					if (!target.exists())
						throw new Exception("directory does not exist: "+ target.getCanonicalPath());
					if (!target.isDirectory())
						throw new Exception("target is no directory: "+ target.getCanonicalPath());
				}
				
				if ( has(flags, SOSFileOperations.RECURSIVE) )
					list = SOSFile.getFilelist(source.getPath(), fileSpec, fileSpecFlags, true);
				else
					list = SOSFile.getFilelist(source.getPath(), fileSpec, fileSpecFlags);
			
			// 2. Quelle ist eine Datei
			} else {
				list = new Vector();
				list.add(source);
			}
			
			File sourceFile, targetFile;
			File dir;
			
			for (int i = 0; i < list.size(); i++) {
				
				sourceFile = (File) list.get(i);
				
				// Zieldatei ermitteln
				if (target != null) {
					if (target.isDirectory()) {
						// Rekursion bercksichtigen
						String root = (source.isDirectory()) ? source.getPath() : source.getParent();
						targetFilename = target.getPath() + sourceFile.getPath().substring(root.length());
					}
					else {
						targetFilename = target.getPath();
					}
				
				// target ist null, d.h. das Quellverzeichnis ist auch das Zielverzeichnis
				}
				else {
					if (source.isDirectory()) {
						// Rekursion bercksichtigen
						String root = (source.isDirectory()) ? source.getPath() : source.getParent();
						targetFilename = source.getPath() + sourceFile.getPath().substring(root.length());
					} else
						targetFilename = source.getParent() + "/" + sourceFile.getName();
				}
				
				targetFile = new File(targetFilename);
				
				// Dateinamen ersetzen
				try {
					
					if ( replace ) {
						
						targetFilename = targetFile.getName(); // bis jetzt noch ursprnglicher Dateiname
						
						targetFilename = replaceGroups(targetFilename, replacing, replacement.split(";"));
						targetFilename = substituteAllDate(targetFilename);
						targetFilename = substituteAllFilename(targetFilename, targetFile.getName());
						
						// should any opening and closing brackets be found in the file name, then this is an error
						Matcher matcher = Pattern.compile("\\[[^]]*\\]").matcher(targetFilename);
						if ( matcher.find() )
							throw new Exception("unsupported file mask found: "+ matcher.group());
						
						targetFile = new File(targetFile.getParent() + "/" + targetFilename); // neuer Dateiname
					}
					
				} catch (Exception re) {
					throw new Exception("replacement error in file " + targetFilename + ": " + re.getMessage());
				}

				// Existieren alle bentigten Elternverzeichnisse?
				dir = new File(targetFile.getParent());

				if (!dir.exists()) {
					if (dir.mkdirs()) {
						log("create directory " + dir.getCanonicalPath(), logger);
					} else
						throw new Exception("cannot create directory " + dir.getCanonicalPath());
				}

				if ( copying ) {
					if ( !copyOneFile(sourceFile, targetFile, overwrite, gracious, logger) ) continue;
				} else if ( renaming ) {
					if ( !renameOneFile(sourceFile, targetFile, overwrite, gracious, logger) ) continue;
				}
				
				nrOfTransferedFiles++;
			}

			if (copying)
				log(nrOfTransferedFiles + " file(s) copied", logger);
			else if (renaming)
				log(nrOfTransferedFiles + " file(s) renamed", logger);
			
			return (nrOfTransferedFiles > 0);
			
		} catch (Exception e) {
			if (copying)
				throw new Exception("error occurred copying file(s): " + e.getMessage());
			else if (renaming)
				throw new Exception("error occurred renaming file(s): " + e.getMessage());
			else return true;
		}
	}
	
	private static String substituteFirstFilename(String targetFilename, String original) throws Exception {
		
		// check for [filename:...]
		Matcher matcher = Pattern.compile("\\[filename:([^\\]]*)\\]").matcher(targetFilename);
		
		if ( matcher.find() ) {
			
			if ( matcher.group(1).equals("") ) {
				targetFilename = targetFilename.replaceFirst("\\[filename:\\]", original);
			}
			else if ( matcher.group(1).equals("lowercase") ) {
				targetFilename = targetFilename.replaceFirst("\\[filename:lowercase\\]", original.toLowerCase());	
			}
			else if ( matcher.group(1).equals("uppercase") ) {
				targetFilename = targetFilename.replaceFirst("\\[filename:uppercase\\]", original.toUpperCase());
			}
		}
		
		return targetFilename;
	}
	
	private static String substituteAllFilename(String targetFilename, String original) throws Exception {
		
		String temp = substituteFirstFilename(targetFilename, original);
		
		while ( !targetFilename.equals(temp) ) {
			targetFilename = temp;
			temp = substituteFirstFilename(targetFilename, original);
		}
		
		return temp;
	}
	
	private static String substituteFirstDate(String targetFilename) throws Exception {
		
		try {
			
			// check for a date format string given in the file mask
			if (targetFilename.matches("(.*)(\\[date:)([^\\]]+)(\\])(.*)")) {
				
				int posBegin = targetFilename.indexOf("[date:");
				if (posBegin > -1) {
					int posEnd = targetFilename.indexOf("]", posBegin + 6);
					if (posEnd > -1) {
						targetFilename = ((posBegin > 0) ? targetFilename.substring(
								0, posBegin) : "")
								+ SOSDate.getCurrentTimeAsString(targetFilename.substring(posBegin + 6, posEnd))
								+ ((targetFilename.length() > posEnd) ? targetFilename.substring(posEnd + 1) : "");
					}
				}
			}
			
			return targetFilename;
			
		} catch (Exception e) {
			throw new Exception("error substituting [date]: "+e.getMessage());
		}
	}
	
	
	private static String substituteAllDate(String targetFilename) throws Exception {
		
		String temp = substituteFirstDate(targetFilename);
		
		while ( !targetFilename.equals(temp) ) {
			targetFilename = temp;
			temp = substituteFirstDate(targetFilename);
		}
		
		return temp;
	}
	
	
	private static String substituteFirstDirectory(String target, String source) throws Exception {
		
		try {
			
			File sourceFile = new File(source);
			if ( !sourceFile.isDirectory() ) {
				source = sourceFile.getParent();
			}
			
			// normalisieren
			source = source.replaceAll("\\\\", "/");
			target = target.replaceAll("\\\\", "/");
			
			
			Pattern p = Pattern.compile("\\[directory:(-[\\d]+|[\\d]*)\\]");
			Matcher m = p.matcher(target);
			
			if ( m.find() ) {
				
				String substitute = "";
				
				if ( m.group(1).length()==0 || m.group(1).equals("0") || m.group(1).equals("-0") ) {
					substitute = source;
				}
				else {
					
					int depth = Integer.valueOf(m.group(1)).intValue();
					StringTokenizer st = new StringTokenizer(source,"/");
					int absDepth = ( depth<0 ) ? -depth : depth;
					
					if ( absDepth >= st.countTokens() ) {
						substitute = source;
					}
					else {
						String[] dirs = new String[st.countTokens()];
						int n = 0;
						while ( st.hasMoreTokens() )
							dirs[n++] = st.nextToken();
						
						if ( depth>0 )
							while ( depth>0 ) substitute = dirs[--depth] +"/"+ substitute;
						else
						if ( depth<0 )
							while ( depth<0 ) substitute = substitute + dirs[ dirs.length+depth++ ] +"/";
					}
				}
				
				if ( substitute.charAt(substitute.length()-1)=='/' )
					substitute = substitute.substring(0, substitute.length()-1);
				
				target = target.replaceFirst("\\[directory:[^\\]]*\\]", substitute);
			}
			
			return target;
			
		} catch (Exception e) {
			throw new Exception("error substituting [directory]: "+e.getMessage());
		}
	}
	
	
	private static String substituteAllDirectory(String target, String source) throws Exception {
		
		String temp = substituteFirstDirectory(target, source);
		
		while ( !target.equals(temp) ) {
			target = temp;
			temp = substituteFirstDirectory(target, source);
		}
		
		return temp;
	}
	

	/**
   * Replaces all groups if a string with given replacements.
   * If only group 0 is existing (that means the entire match) it will be replaced
   * with the first replacement. Otherwise each capturing group will be replaced accordingly.
   * 
   * @param input
   *          Inputstring
   * @param replacing
   *          Pattern
   * @param replacements
   *          Semikolon-getrennte Ersetzungsstrings
   * @return Result
   * @throws Exception
   */
	public static String replaceGroups(String input, String replacing, String replacements) throws Exception {
		
		if (replacements == null || replacements.length() == 0)
			throw new Exception("replacements missing: 0 replacements defined");
		
		return replaceGroups(input, replacing, replacements.split(";"));
	}
	
	
	/**
	 * Replaces all groups if a string with given replacements.
	 * If only group 0 is existing (that means the entire match) it will be replaced
	 * with the first replacement. Otherwise each capturing group will be replaced accordingly.
	 * 
	 * @param input
	 *          Inputstring
	 * @param replacing
	 *          Pattern
	 * @param replacements
	 *          Array von Ersetzungsstrings
	 * @return Result
	 * @throws Exception
	 */
	public static String replaceGroups(String input, String replacing,
			String[] replacements) throws Exception {
		
		String result = "";
		
		if (replacements == null || replacements.length == 0)
			throw new Exception("replacements missing: 0 replacements defined");
		
		Pattern p = Pattern.compile(replacing);
		Matcher m = p.matcher(input);
		
		if (!m.find())
			return input;
		
		if (replacements.length < m.groupCount())
			throw new Exception("replacements missing: " + replacements.length
					+ " replacement(s) defined but " + m.groupCount() + " groups found");
		
		// gesamte Gruppe 0 ersetzen
		if (m.groupCount() == 0) {
			result = input.substring(0, m.start()) + replacements[0]
					+ input.substring(m.end());
			
		} else {
			
			int index = 0;
			
			for (int i = 1; i <= m.groupCount(); i++) {
				if (m.start(i) >= 0) {
					result += input.substring(index, m.start(i)) + replacements[i - 1];
					index = m.end(i);
				}
			}

			result += input.substring(index);
		}

		return result;
	}

	private static boolean copyOneFile(File source, File target, boolean overwrite, boolean gracious, SOSLogger logger) throws Exception {
		
		boolean rc = false;
		
		if (source.equals(target))
			throw new Exception("cannot copy file to itself: " + source.getCanonicalPath());
		
		if ( overwrite || !target.exists() ) {
			long modificationDate = source.lastModified();
			rc = SOSFile.copyFile(source, target);
			target.setLastModified(modificationDate);
			log("copy " + source.getPath() + " to " + target.getPath(), logger);
			return rc;
		}
		else if ( !gracious ) {
			throw new Exception("file already exists: " + target.getCanonicalPath());
		}
		else {
			log("file already exists: " + target.getCanonicalPath(), logger);
			return rc;
		}
	}
	
	private static boolean renameOneFile(File source, File target,
			boolean overwrite, boolean gracious, SOSLogger logger) throws Exception {
        
		if (source.equals(target))
			throw new Exception("cannot rename file to itself: "
					+ source.getCanonicalPath());

		if (!overwrite && target.exists()) {

            if (!gracious) {
                throw new Exception("file already exists: " + target.getCanonicalPath());
            } else {
                log("file already exists: " + target.getCanonicalPath(), logger);
                return false;
            }
		} else {

            if (target.exists()) {
				if (!target.delete())
					throw new Exception("cannot overwrite " + target.getCanonicalPath());
			}
			
			if (!source.renameTo(target)) {
				throw new Exception("cannot rename file from "
						+ source.getCanonicalPath() + " to " + target.getCanonicalPath());
			}
			else {
				log("rename " + source.getPath() + " to " + target.getPath(), logger);
				return true;
			}
		}
	}
	
	
	private static void log(String msg, SOSLogger logger) {

		try {
			if (logger != null)
				logger.info(msg);
		} catch (Exception e) {
		}
	}

	private static void log_debug1(String msg, SOSLogger logger) {

		try {
			if (logger != null)
				logger.debug1(msg);
		} catch (Exception e) {
		}
	}

	/**
	 * Tests if a flag is set or not
	 */
	private static boolean has(int flags, int f) {
		return (flags & f) > 0;
	}
	
	/**
	 * Method for invoking a method of this class by use of the Reflection API
	 * 
	 * @param methodname
	 * @param argtypes
	 * @param args
	 * @throws Exception
	 */
	public static void callMethod(String methodname, Class[] argtypes, Object[] args) throws Exception {
		
		Method method = null;
		
		try {

			if (argtypes.length != args.length)
				throw new Exception("different array lengths: " + argtypes.length
						+ " argument types but " + args.length + " arguments");

			try {
				method = Class.forName("sos.util.SOSFileOperations").getMethod(
						methodname, argtypes);
			} catch (NoSuchMethodException nsme) {
				throw new Exception("method does not exist: " + nsme.getMessage());
			}

		} catch (Exception e) {
			throw new Exception("callMethod: " + e.getMessage());
		}

		try {
			method.invoke(null, args);
		} catch (Exception x) {
			if (x.getMessage() == null)
				throw new Exception(x.getCause().getMessage());
			else
				throw new Exception("callMethod: " + x.getMessage());
		}
	}
	
	
	/**
	 * Liefert einen neuen Dateinamen
	 */
	public static String getReplacementFilename(String input, String replacing,
			String replacements) throws Exception {
		
		String targetFilename = input;
		try {
			targetFilename = replaceGroups(targetFilename, replacing,
					replacements.split(";"));
			
			// check for a date format string given in the file mask
			if (targetFilename.matches("(.*)(\\[date\\:)([^\\]]+)(\\])(.*)")) {
				int posBegin = targetFilename.indexOf("[date:");
				if (posBegin > -1) {
					int posEnd = targetFilename.indexOf("]", posBegin + 6);
					if (posEnd > -1) {
						targetFilename = ((posBegin > 0) ? targetFilename.substring(
								0, posBegin) : "")
								+ SOSDate.getCurrentTimeAsString(targetFilename.substring(posBegin + 6, posEnd))
								+ ((targetFilename.length() > posEnd) ? targetFilename.substring(posEnd + 1) : "");
					}
				}
			}
			
			// check for [filename:]
			if (targetFilename.matches("(.*)(\\[filename\\:\\])(.*)")) {
				targetFilename = targetFilename.replaceAll("\\[filename\\:\\]",input);
			}
			
			// should any opening and closing brackets be found in the file name, then this is an error
			Matcher m = Pattern.compile("\\[[^\\]]*\\]").matcher(targetFilename);
			if ( m.find() ) throw new Exception("unsupported file mask found:"+ m.group());
			
			return targetFilename;
			//targetFile = new File(targetFile.getParent() + "/" + targetFilename);
			//return targetFile.getName();
		} catch (Exception e) {
			throw new Exception("getReplacementFilename: " + e.getMessage());
		}
	}

	
	private static boolean wipe(File file, SOSLogger logger) {
        try {     
        	RandomAccessFile rwFile = new RandomAccessFile(file, "rw");
            
        	byte[] bytes = new byte[(int)(rwFile.length())];
            int i = 0;
            while ( (bytes[i++] = (byte)rwFile.read()) != -1 );
            rwFile.seek(0);
            for ( i=0; i < bytes.length; i++)
              bytes[i] = (byte)((Math.random()*10)%9);
            rwFile.write(bytes);
            rwFile.close();
            logger.debug9("Deleting file");
            boolean rc = file.delete();
            logger.debug9("rc: "+rc);
            return rc;
        } catch(Exception e) {
        	try{
            logger.warn("Failed to wipe file: "+e);
        	} catch (Exception ex){}
        	return false;
        }
	}
	
	/*
	public static void main(String[] args) {
		
		try {
			
			SOSLogger logger = new SOSStandardLogger(1);
			
			//String source = "files/";
			
			//String target = "target/[date:yyyyMMdd]/[directory:2]";
			//String target = "target/";
			
			//renameFile(source, ".*dat$", 0, Pattern.CASE_INSENSITIVE, "(ab)_c[\\d]d_(fg)", "first;second", logger);
			
			//removeFile("K:/scheduler.test/files/file_operations/copy/12/out", ".*", RECURSIVE | REMOVE_DIR, logger);
			
			//renameFile("files/", ".*", 0, Pattern.CASE_INSENSITIVE, ".*", "[filename:]y", logger);
			
			// /aaa/bbb/ccc/datei.txt sollte umbenannt werden nach /aaa/bbb/ccc/20070101/datei_20070101.txt
			//copyFile("files/aaa/bbb/ccc/", "[directory:]/[date:yyyyMMdd]/", "datei.txt", CREATE_DIR, Pattern.CASE_INSENSITIVE, "[.]txt", "_[date:yyyyMMdd].txt", logger);
			
			copyFile("files/file_operations/copy/12/in/", "[directory:4]/out/[date:yyyyMMdd]/[directory:-3]/", ".*", CREATE_DIR, Pattern.CASE_INSENSITIVE, logger);
			
			//copyFile(source, target, ".*", CREATE_DIR|RECURSIVE, Pattern.CASE_INSENSITIVE, ".*", "[filename:]", logger);
			//copyFile(source, target, "da.*", RECURSIVE & 0, Pattern.CASE_INSENSITIVE, ".*X.*", "[date:yyyyMMdd]_[filename:uppercase]_[filename:lowercase]_[filename:]_[date:yyyy]", logger);
			
		} catch (Exception e) {
			System.err.println("main: " + e.getMessage());
		}
	}*/
	
}
